using System;
using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Threading;
using MVR;
using MeshVR;
using AssetBundles;
using System.Linq;
using Battlehub.RTCommon;
using GPUTools.Physics.Scripts.Behaviours;
using System.Runtime.CompilerServices;
using Leap;
using Microsoft.Win32;
using System.Security.Authentication;
using Leap.Unity.Infix;
using Leap.Unity;

namespace Stopper {
	public class ShoeHelper : MVRScript {
		DAZCharacterSelector characterSelector;
		
		List<ConfigurableJoint> leftToeJoints = new List<ConfigurableJoint>();
		List<ConfigurableJoint> rightToeJoints = new List<ConfigurableJoint>();
		// List<ConfigurableJoint> toeJoints = new List<ConfigurableJoint>();
		
		string selectedClothingUid;
		DAZClothingItem shoeClothingItem;
		
		float shoeFootDriveTarget = -45f;
		float shoeToeDriveTarget = 45f;

		bool isFirstSync = true;
		float originalShoeFootDriveTarget = 0f;
		float originalShoeToeDriveTarget = 0f;
		

		DAZBone lFoot;
		DAZBone rFoot;

		DAZBone lToe;
		DAZBone rToe;

		JSONStorableBool adjustController;

		bool isClothingPlugin = false;

		BoxCollider soleCollider;


		Dictionary<GameObject, BoxCollider> colliderPreviews = new Dictionary<GameObject, BoxCollider>();

		List<GameObject> createdObjects = new List<GameObject>();

		JSONStorableBool showColliders;

		ShoeCollider leftHeelCollider;
		ShoeCollider leftToeCollider;

		ShoeCollider rightHeelCollider;
		ShoeCollider rightToeCollider;

		List<ShoeCollider> shoeColliders = new List<ShoeCollider>();
		
        MVRScriptExtensions.VerticalLayoutPanel mainPanel;

		UIDynamicPopup selectShoePopup;
		
		
		JSONStorableBool leftEnabled;
		JSONStorableBool rightEnabled;

		UIDynamicToggle leftEnabledToggle;
		UIDynamicToggle rightEnabledToggle;

		UIDynamicToggle adjustControllerRotation;
		UIDynamicSlider toeRotationTarget;
		UIDynamicSlider footRotationTarget;


		UIDynamicSlider toeStiffness;
		UIDynamicSlider toeDamper;
		UIDynamicSlider toeMaxForce;

		// Colliders
		UIDynamicToggle showCollidersToggle;

		//Heel
		Vector3 heelColliderDimensions;
		Vector3 heelColliderCenter;
		Vector3 heelColliderRotation;

		JSONStorableBool heelColliderEnabledBool;
		UIDynamicToggle heelColliderEnabled;

		UIDynamicSlider heelColliderLength;
		UIDynamicSlider heelColliderWidth;

		UIDynamicSlider heelColliderXOffset;
		UIDynamicSlider heelColliderYOffset;
		UIDynamicSlider heelColliderZOffset;

		UIDynamicSlider heelColliderXAngle;
		UIDynamicSlider heelColliderYAngle;

		//Toe
		Vector3 toeColliderDimensions;
		Vector3 toeColliderCenter;
		Vector3 toeColliderRotation;

		JSONStorableStringChooser toeParentChooser;
		Transform leftToeSelectedParent;
		Transform rightToeSelectedParent;

		JSONStorableBool toeColliderEnabledBool;
		UIDynamicToggle toeColliderEnabled;

		UIDynamicPopup toeColliderParent;

		UIDynamicSlider toeColliderXScale;
		UIDynamicSlider toeColliderYScale;
		UIDynamicSlider toeColliderZScale;

		UIDynamicSlider toeColliderXOffset;
		UIDynamicSlider toeColliderYOffset;
		UIDynamicSlider toeColliderZOffset;

		UIDynamicSlider toeColliderXAngle;
		UIDynamicSlider toeColliderYAngle;
		UIDynamicSlider toeColliderZAngle;

		public class ShoeCollider {
			public GameObject gameObject;
			public Rigidbody rigidbody;
			public ConfigurableJoint joint;

			public Transform parent;

			public void SetTransform(Vector3 scale, Vector3 offset, Vector3 rotation) {
				var parent = joint.connectedBody;
				joint.connectedBody = null;
				rigidbody.transform.localScale = scale;
				rigidbody.transform.localPosition = offset;
				rigidbody.transform.localRotation = Quaternion.Euler(rotation);

				joint.connectedBody = parent;
			}

			public void UpdateParent(Transform selectedParent) {
				if(selectedParent != null && selectedParent != this.parent) {
					gameObject.transform.SetParent(selectedParent, false);
					joint.connectedBody = null;
					joint.connectedBody = selectedParent.GetComponent<Rigidbody>();
					this.parent = selectedParent;
				}
			}

			public void SyncShowCollider(bool show) {
				foreach(Renderer r in gameObject.GetComponentsInChildren<Renderer>(true)) {
					r.enabled = show;
				}
			}
		}
		
		
		public override void Init() {
			WalkAndGetParentItems();
			// We found a clothing item in our hierarchy. We're probably being used as a clothing plugin.
			// Otherwise something very weird has happened...			
			isClothingPlugin = shoeClothingItem != null;
	
			characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
			
			var physicsModel = containingAtom.GetStorableByID("PhysicsModel");
			foreach(Rigidbody rb in physicsModel.GetComponentsInChildren<Rigidbody>()){
				if(rb.name.Contains("lSmallToe") || rb.name.Contains("lBigToe") ){
					// There may be multple joints, but this seems to get the right one
					leftToeJoints.Add(rb.GetComponent<ConfigurableJoint>());
				}
				if(rb.name.Contains("rSmallToe") || rb.name.Contains("rBigToe") ){
					// There may be multple joints, but this seems to get the right one
					rightToeJoints.Add(rb.GetComponent<ConfigurableJoint>());
				}
			}

			lFoot = containingAtom.GetStorableByID("lFoot") as DAZBone;
			rFoot = containingAtom.GetStorableByID("rFoot") as DAZBone;

			lToe = containingAtom.GetStorableByID("lToe") as DAZBone;
			rToe = containingAtom.GetStorableByID("rToe") as DAZBone;

			if(!isClothingPlugin) {
				JSONStorableStringChooser clothingChooser = new JSONStorableStringChooser(
						"High Heel Clothing", 
						containingAtom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList(),
						null, 
						"High Heel Clothing", 
						(string uid) => {
							if(shoeClothingItem != null){
								DisconnectClothingItem(shoeClothingItem);
							}
							
							shoeClothingItem = characterSelector.GetClothingItem(uid);

							
							if(shoeClothingItem != null && !shoeClothingItem.name.Contains("Harli")) {
								SyncShoeSettings(shoeClothingItem);
								SyncDriveControllers(shoeClothingItem);
							} else {
								SuperController.LogMessage($"Could not find Clothing {selectedClothingUid}");	
							}
						});
						
				RegisterStringChooser(clothingChooser);
				selectShoePopup = CreateScrollablePopup(clothingChooser, false);
				selectShoePopup.popup.onOpenPopupHandlers += () => {
					clothingChooser.choices = containingAtom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList();
				};
			}


			leftEnabled = new JSONStorableBool("Left Foot Enabled", true);
			RegisterBool(leftEnabled);
			leftEnabledToggle = CreateToggle(leftEnabled, false);
			leftEnabledToggle.toggle.onValueChanged.AddListener((val) => {
				OnDisable();
				OnEnable();
			});

			rightEnabled = new JSONStorableBool("Right Foot Enabled", true);
			RegisterBool(rightEnabled);
			rightEnabledToggle = CreateToggle(rightEnabled, false);
			rightEnabledToggle.toggle.onValueChanged.AddListener((val) => {
				OnDisable();
				OnEnable();
			});



			adjustController = new JSONStorableBool("Adjust Controller Rotation", true);
			RegisterBool(adjustController);
			adjustControllerRotation = CreateToggle(adjustController, false);
			adjustControllerRotation.toggle.onValueChanged.AddListener((val) => {
				if(val) {
					AdjustControllerRotation(shoeFootDriveTarget);
				} else {
					AdjustControllerRotation(0f);
				}
			});

		
			toeRotationTarget = AddSlider("toeRotationTarget", "Toe Rotation Target", shoeToeDriveTarget, -75f, 75f, (float val) => {
				shoeToeDriveTarget = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			footRotationTarget = AddSlider("footRotationTarget", "Foot Rotation Target", shoeFootDriveTarget, -65f, 65f, (float val) => {
				shoeFootDriveTarget = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			toeStiffness = AddSlider("toeStiffness", "Toe Stiffness", 10f, 0f, 2000f, (float val) => {
				if(leftEnabled.val){
					foreach(var joint in leftToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.positionSpring = val;
						joint.slerpDrive = slerpDrive;
					}
				}
				if(rightEnabled.val){
					foreach(var joint in rightToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.positionSpring = val;
						joint.slerpDrive = slerpDrive;
					}
				}
			}, true);
			
			toeDamper = AddSlider("toeDamper", "Toe Damper", 10f, 0f, 2000f, (float val) => {
				if(leftEnabled.val){
					foreach(var joint in leftToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.positionDamper = val;
						joint.slerpDrive = slerpDrive;
					}
				}
				if(rightEnabled.val){
					foreach(var joint in rightToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.positionDamper = val;
						joint.slerpDrive = slerpDrive;
					}
				}
			}, true);
			
			toeMaxForce = AddSlider("toeMaxForce", "Toe Max Force", 10f, 0f, 2000f, (float val) => {
				if(leftEnabled.val){
					foreach(var joint in leftToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.maximumForce = val;
						joint.slerpDrive = slerpDrive;
					}
				}
				if(rightEnabled.val){
					foreach(var joint in rightToeJoints){
						var slerpDrive = joint.slerpDrive;
						slerpDrive.maximumForce = val;
						joint.slerpDrive = slerpDrive;
					}
				}
			}, true);


			showColliders = new JSONStorableBool("Show Colliders", false);
			showColliders.isStorable = false;
			RegisterBool(showColliders);
			showCollidersToggle = CreateToggle(showColliders, false);
			showCollidersToggle.toggle.onValueChanged.AddListener((val) => {
				foreach(var sc in shoeColliders) {
					sc.SyncShowCollider(val);
				}
			});

			// Heel Colliders
			heelColliderEnabledBool = new JSONStorableBool("heelColliderEnabled", true);
			RegisterBool(heelColliderEnabledBool);
			heelColliderEnabled = CreateToggle(heelColliderEnabledBool, false);
			heelColliderEnabled.label = "Enabled";
			heelColliderEnabled.toggle.onValueChanged.AddListener((val) => {
				SyncShoeSettings(shoeClothingItem);
			});

			heelColliderLength = AddSlider("heelColliderLength", "Length", 0.6f, 0f, 2f, (float val) => {
				heelColliderDimensions.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			heelColliderWidth = AddSlider("heelColliderWifth", "Width", 1f, 0f, 2f, (float val) => {
				heelColliderDimensions.x = val;
				heelColliderDimensions.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			heelColliderXOffset = AddSlider("heelColliderXOffset", "X Offset", 0.00f, -0.2f, 0.2f, (float val) => {
				heelColliderCenter.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			heelColliderYOffset = AddSlider("heelColliderYOffset", "Y Offset", -0.06f, -0.2f, 0.2f, (float val) => {
				heelColliderCenter.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			heelColliderZOffset = AddSlider("heelColliderZOffset", "Z Offset", 0.03f, -0.2f, 0.2f, (float val) => {
				heelColliderCenter.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			heelColliderXAngle = AddSlider("heelColliderXRotation", "X Rotation", 115f, -180f, 180f, (float val) => {
				heelColliderRotation.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			heelColliderYAngle = AddSlider("heelColliderYRotation", "Y Rotation", 0f, -180f, 180f, (float val) => {
				heelColliderRotation.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			
			// Toe Colliders
			toeParentChooser = new JSONStorableStringChooser(
						"toeColliderParent", 
						new List<string>() {"Toe", "Foot"},
						"Toe", 
						"Parent Bone", 
						(string parent) => {
							SyncShoeSettings(shoeClothingItem);
						});
						
			RegisterStringChooser(toeParentChooser);
			toeColliderParent = CreateScrollablePopup(toeParentChooser, false);
				

			toeColliderEnabledBool = new JSONStorableBool("toeColliderEnabled", true);
			RegisterBool(toeColliderEnabledBool);
			toeColliderEnabled = CreateToggle(toeColliderEnabledBool, false);
			toeColliderEnabled.label = "Enabled";
			toeColliderEnabled.toggle.onValueChanged.AddListener((val) => {
				SyncShoeSettings(shoeClothingItem);
			});

			toeColliderXScale = AddSlider("toeColliderXScale", "X Scale", 0.15f, -0.2f, 0.2f, (float val) => {
				toeColliderDimensions.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderYScale = AddSlider("toeColliderYScale", "Y Scale", 0.03f, -0.2f, 0.2f, (float val) => {
				toeColliderDimensions.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderZScale = AddSlider("toeColliderZScale", "Z Scale", 0.15f, -0.2f, 0.2f, (float val) => {
				toeColliderDimensions.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderXOffset = AddSlider("toeColliderXOffset", "X Offset", 0.00f, -0.2f, 0.2f, (float val) => {
				toeColliderCenter.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderYOffset = AddSlider("toeColliderYOffset", "Y Offset", -0.02f, -0.2f, 0.2f, (float val) => {
				toeColliderCenter.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderZOffset = AddSlider("toeColliderZOffset", "Z Offset", 0.03f, -0.2f, 0.2f, (float val) => {
				toeColliderCenter.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);


			toeColliderXAngle = AddSlider("toeColliderXRotation", "X Rotation", -20f, -180f, 180f, (float val) => {
				toeColliderRotation.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderYAngle = AddSlider("toeColliderYRotation", "Y Rotation", 0f, -180f, 180f, (float val) => {
				toeColliderRotation.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			toeColliderZAngle = AddSlider("toeColliderZRotation", "Z Rotation", 0f, -180f, 180f, (float val) => {
				toeColliderRotation.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
		}

		public void Start() {
			InitMainUI();

			if(toeParentChooser == null || toeParentChooser.val == "Foot") {
				leftToeSelectedParent = lFoot.transform;
				rightToeSelectedParent = rFoot.transform;
			} else {
				leftToeSelectedParent = lToe.transform;
				rightToeSelectedParent = rToe.transform;
			}

			foreach(var entry in this.floatParams) {
				var floatParam = entry.Value;
				floatParam.setCallbackFunction(floatParam.val);
			}

			if(isClothingPlugin) {
				OnEnable();
			}

			StartCoroutine(AddColliderAtom("Capsule", lFoot.transform, DAZClothingItem.ControllerType.LeftFoot));
			StartCoroutine(AddColliderAtom("Capsule", rFoot.transform, DAZClothingItem.ControllerType.RightFoot));

			StartCoroutine(AddColliderAtom("Cube", leftToeSelectedParent, DAZClothingItem.ControllerType.LeftToe));
			StartCoroutine(AddColliderAtom("Cube", rightToeSelectedParent, DAZClothingItem.ControllerType.RightToe));
		}

		public void InitMainUI() {
            if(mainPanel == null) {


                var mainUI = this.rightUIContent.transform.parent;
                mainPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("MainPanel", mainUI);

				var selectionPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("SelectionPanel", mainPanel);
				var posePanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("PosePanel", mainPanel);
				var toePanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("ToePanel", mainPanel);
				var collidersPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("CollidersPanel", mainPanel);
		
				if(!isClothingPlugin) {
					this.AddToPanel(selectShoePopup, selectionPanel);
				}

				this.AddToPanel(leftEnabledToggle, selectionPanel);
				this.AddToPanel(rightEnabledToggle, selectionPanel);

				// Pose Controls
				var poseControlsHeader = CreateHeader("Pose Controls", posePanel);

				this.AddToPanel(adjustControllerRotation, posePanel);
				
				var poseSliders = this.AddUIElement<MVRScriptExtensions.VerticalSplitLayoutPanel>("PoseSliders", posePanel);
				this.AddToPanel(toeRotationTarget, poseSliders.leftPanel);
				this.AddToPanel(footRotationTarget, poseSliders.rightPanel);


				// toe Controls
				var toeControlsHeader = CreateHeader("Toe Controls", toePanel);

				var toeSliders = this.AddUIElement<MVRScriptExtensions.VerticalSplitLayoutPanel>("ToeSliders", toePanel);
				this.AddToPanel(toeStiffness, toeSliders.leftPanel);
				this.AddToPanel(toeDamper, toeSliders.leftPanel);
				this.AddToPanel(toeMaxForce, toeSliders.rightPanel);

				// Colldier controls
				var collidersHeader = CreateHeader("Colliders", collidersPanel);
				this.AddToPanel(showCollidersToggle, collidersPanel);

				var collidersSliders = this.AddUIElement<MVRScriptExtensions.VerticalSplitLayoutPanel>("ColliderSliders", collidersPanel);

				var heelHeader = this.AddTextBox("Heel", collidersSliders.leftPanel);

				this.AddToPanel(heelColliderEnabled, collidersSliders.leftPanel);
				
				this.AddToPanel(heelColliderLength, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderWidth, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderXOffset, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderYOffset, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderZOffset, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderXAngle, collidersSliders.leftPanel);
				this.AddToPanel(heelColliderYAngle, collidersSliders.leftPanel);

				var toeHeader = this.AddTextBox("Toe", collidersSliders.rightPanel);
				this.AddToPanel(toeColliderEnabled, collidersSliders.rightPanel);

				this.AddToPanel(toeColliderParent, collidersSliders.rightPanel);

				this.AddToPanel(toeColliderXScale, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderYScale, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderZScale, collidersSliders.rightPanel);

				this.AddToPanel(toeColliderXOffset, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderYOffset, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderZOffset, collidersSliders.rightPanel);

				this.AddToPanel(toeColliderXAngle, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderYAngle, collidersSliders.rightPanel);
				this.AddToPanel(toeColliderZAngle, collidersSliders.rightPanel);
            }
        }

		private MVRScriptExtensions.TextBox CreateHeader(string text, MVRScriptExtensions.LayoutPanel panel) {
			var header = this.AddTextBox(text, panel);
			header.UIText.alignment = TextAnchor.UpperLeft;
			header.UIText.fontStyle = FontStyle.Bold;

			return header;
		}

		private void WalkAndGetParentItems(){
			GameObject parent = this.gameObject;
			while(containingAtom == null || parent != containingAtom.gameObject){
				if(parent.transform.parent == null){
					return;
				}
				
				if(shoeClothingItem == null){
					shoeClothingItem = parent.GetComponent<DAZClothingItem>();
				}

				if(containingAtom == null) {
                	containingAtom = parent.GetComponent<Atom>();
				}
				parent = parent.transform.parent.gameObject;
			}
		}

		private IEnumerator AddColliderAtom(string shape, Transform parent, DAZClothingItem.ControllerType controllerType) {
			AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_sha", shape, typeof(GameObject));
			if (request == null)
			{
				yield break;
			}
			yield return base.StartCoroutine(request);
			
			GameObject go = request.GetAsset<GameObject>();

			if(go == null) {
				SuperController.LogError($"Could not initialize collider for {controllerType.ToString()}");
			}
			GameObject colliderObject = UnityEngine.Object.Instantiate<Transform>(go.transform).gameObject;


			ShoeCollider collider = new ShoeCollider();
			shoeColliders.Add(collider);
			collider.gameObject = colliderObject;

			createdObjects.Add(colliderObject.gameObject);

			Atom colliderAtom = colliderObject.GetComponent<Atom>();
			DestroyImmediate(colliderAtom);

			// Set up initial position
			collider.parent = parent;
			colliderObject.transform.parent = parent;
			colliderObject.transform.position = parent.position;
			colliderObject.transform.rotation = parent.rotation;

			colliderObject.transform.localPosition = Vector3.zero;
			colliderObject.transform.localRotation = Quaternion.identity;

			// Remove the free controller
			foreach(FreeControllerV3 controller in colliderObject.GetComponentsInChildren<FreeControllerV3>(true)){
				DestroyImmediate(controller.gameObject);
			}

			// Set the physics objects
			Rigidbody rb = colliderObject.GetComponentInChildren<Rigidbody>(true);
			collider.rigidbody = rb;
			rb.useGravity = false;
	
			ConfigurableJoint joint = rb.GetComponent<ConfigurableJoint>();
			collider.joint = joint;

			joint.connectedBody = parent.GetComponent<Rigidbody>();
			joint.xMotion = ConfigurableJointMotion.Locked;
			joint.yMotion = ConfigurableJointMotion.Locked;
			joint.zMotion = ConfigurableJointMotion.Locked;
			joint.angularXMotion = ConfigurableJointMotion.Locked;
			joint.angularYMotion = ConfigurableJointMotion.Locked;
			joint.angularZMotion = ConfigurableJointMotion.Locked;
			joint.rotationDriveMode =  RotationDriveMode.XYAndZ;

			foreach(Collider c1 in colliderObject.GetComponentsInChildren<Collider>(true)) {
				c1.material.staticFriction = 0f;
				foreach(Collider c2 in parent.GetComponentsInChildren<Collider>(true)) {
					UnityEngine.Physics.IgnoreCollision(c1, c2);
				}
			}

			foreach(DelayedMeshRenderer renderer in rb.GetComponentsInChildren<DelayedMeshRenderer>(true)) {
				renderer.enabled = false;
			}

			foreach(Renderer renderer in rb.GetComponentsInChildren<Renderer>(true)){
				renderer.material = new Material(Shader.Find("Battlehub/RTGizmos/Handles"));

				renderer.material.SetFloat("_Offset", 1f);
				renderer.material.SetFloat("_MinAlpha", 1f);
				Color color = Color.green;
				color.a = 0.3f;
				renderer.material.color = color;

				renderer.enabled = false;
			}

			foreach(Collider c in colliderObject.GetComponentsInChildren<Collider>(true)) {
				c.sharedMaterial.dynamicFriction = 0.1f;
				c.sharedMaterial.staticFriction = 0.01f;
				c.sharedMaterial.frictionCombine = PhysicMaterialCombine.Minimum;
			}

			collider.gameObject.SetActive(false);

			switch (controllerType) {
				case DAZClothingItem.ControllerType.LeftFoot:
					leftHeelCollider = collider;
					break;
				case DAZClothingItem.ControllerType.LeftToe:
					leftToeCollider = collider;
					break;
				case DAZClothingItem.ControllerType.RightFoot:
					rightHeelCollider = collider;
					break;
				case DAZClothingItem.ControllerType.RightToe:
					rightToeCollider = collider;
					break;
			}

			SyncShoeSettings(shoeClothingItem);
		}

		public void Update() {
			if(this.showColliders != null && showColliders.val) {
				foreach(var entry in colliderPreviews) {
					entry.Key.transform.localScale = entry.Value.size;
					entry.Key.transform.localPosition = entry.Value.center;
				}
			}
		}
		
		private void SyncShoeSettings(DAZClothingItem shoeClothingItem){
			if(shoeClothingItem == null) {
				return;
			}

			if(isFirstSync) {
				originalShoeFootDriveTarget = shoeClothingItem.driveXAngleTarget;
				originalShoeToeDriveTarget = shoeClothingItem.drive2XAngleTarget;
				isFirstSync = false;
			}

			shoeClothingItem.driveXAngleTarget = shoeToeDriveTarget;
			shoeClothingItem.drive2XAngleTarget = shoeFootDriveTarget;

			if(rightEnabled.val) {
				// SuperController.LogMessage("Right enabled");
				// shoeClothingItem.colliderTypeRight = DAZClothingItem.ColliderType.Shoe;
				shoeClothingItem.driveXAngleTargetController1Type = DAZClothingItem.ControllerType.RightToe;
				shoeClothingItem.drive2XAngleTargetController1Type = DAZClothingItem.ControllerType.RightFoot;
			} 

			if(leftEnabled.val) {
				// shoeClothingItem.colliderTypeLeft = DAZClothingItem.ColliderType.Shoe;
				shoeClothingItem.driveXAngleTargetController2Type = DAZClothingItem.ControllerType.LeftToe;
				shoeClothingItem.drive2XAngleTargetController2Type = DAZClothingItem.ControllerType.LeftFoot;
			}

			SyncDriveControllers(shoeClothingItem);
			
			RefreshDazClothingItem();

			if(adjustController.val){
				AdjustControllerRotation(shoeFootDriveTarget);
			}

			StartCoroutine(SyncColliders());
		}

		private IEnumerator SyncColliders() {
			if(toeParentChooser == null || toeParentChooser.val == "Foot") {
				leftToeSelectedParent = lFoot.transform;
				rightToeSelectedParent = rFoot.transform;
			} else {
				leftToeSelectedParent = lToe.transform;
				rightToeSelectedParent = rToe.transform;
			}

		
			yield return new WaitForFixedUpdate();
			if(rightEnabled.val) {
				if(rightHeelCollider != null) {
					rightHeelCollider.SetTransform(heelColliderDimensions, Vector3.Reflect(heelColliderCenter, Vector3.left), Vector3.Reflect(heelColliderRotation, Vector3.up));
					rightHeelCollider.gameObject.SetActive(heelColliderEnabledBool.val);
				}

				if(rightToeCollider != null) {
					rightToeCollider.SetTransform(toeColliderDimensions, Vector3.Reflect(toeColliderCenter, Vector3.left), new Vector3(toeColliderRotation.x, -toeColliderRotation.y, -toeColliderRotation.z));
					rightToeCollider.UpdateParent(rightToeSelectedParent);
					rightToeCollider.gameObject.SetActive(toeColliderEnabledBool.val);
				}
			}

			if(leftEnabled.val) {
				if(leftHeelCollider != null) {
					leftHeelCollider.SetTransform(heelColliderDimensions, heelColliderCenter, heelColliderRotation);
					leftHeelCollider.gameObject.SetActive(heelColliderEnabledBool.val);
				}

				if(leftToeCollider != null) {
					leftToeCollider.SetTransform(toeColliderDimensions, toeColliderCenter, toeColliderRotation);
					leftToeCollider.UpdateParent(leftToeSelectedParent);
					leftToeCollider.gameObject.SetActive(toeColliderEnabledBool.val);
				}
			}
			
			yield break;
		}

		private void RefreshDazClothingItem() {
			if (shoeClothingItem != null) {
				if (rightEnabled.val) {
					if (shoeClothingItem.driveXAngleTargetController1 != null) {
						shoeClothingItem.driveXAngleTargetController1.jointRotationDriveXTarget = shoeClothingItem.driveXAngleTarget;
					}
					if (shoeClothingItem.drive2XAngleTargetController1 != null) {
						shoeClothingItem.drive2XAngleTargetController1.jointRotationDriveXTarget = shoeClothingItem.drive2XAngleTarget;
					}
				}

				if (leftEnabled.val) {
					if (shoeClothingItem.driveXAngleTargetController2 != null) {
						shoeClothingItem.driveXAngleTargetController2.jointRotationDriveXTarget = shoeClothingItem.driveXAngleTarget;
					}

					if (shoeClothingItem.drive2XAngleTargetController2 != null) {
						shoeClothingItem.drive2XAngleTargetController2.jointRotationDriveXTarget = shoeClothingItem.drive2XAngleTarget;
					}
				}
			}
		}

		private void AdjustControllerRotation(float angle){
			if(leftEnabled.val) {
				foreach(var j in lFoot.GetComponents<ConfigurableJoint>()){
					var	controller = j.connectedBody.GetComponent<FreeControllerV3>();
					if(controller != null){
						var targetRotation = j.targetRotation.eulerAngles;
						targetRotation.x = angle;
						j.targetRotation = Quaternion.Euler(targetRotation);
					}
				}
			}

			if(rightEnabled.val) {
				foreach(var j in rFoot.GetComponents<ConfigurableJoint>()){
					var	controller = j.connectedBody.GetComponent<FreeControllerV3>();
					if(controller != null){
						var targetRotation = j.targetRotation.eulerAngles;
						targetRotation.x = angle;
						j.targetRotation = Quaternion.Euler(targetRotation);
					}
				}
			}
		}


		// Copied from DAZCharacterSelector
		private void SyncDriveControllers(DAZClothingItem dci) {

			switch (dci.driveXAngleTargetController1Type) {
				case DAZClothingItem.ControllerType.None:
					dci.driveXAngleTargetController1 = null;
					break;
				case DAZClothingItem.ControllerType.LeftFoot:
					dci.driveXAngleTargetController1 = characterSelector.leftFootController;
					break;
				case DAZClothingItem.ControllerType.RightFoot:
					dci.driveXAngleTargetController1 = characterSelector.rightFootController;
					break;
				case DAZClothingItem.ControllerType.LeftToe:
					dci.driveXAngleTargetController1 = characterSelector.leftToeController;
					break;
				case DAZClothingItem.ControllerType.RightToe:
					dci.driveXAngleTargetController1 = characterSelector.rightToeController;
					break;
			}

			switch (dci.drive2XAngleTargetController1Type) {
				case DAZClothingItem.ControllerType.None:
					dci.drive2XAngleTargetController1 = null;
					break;
				case DAZClothingItem.ControllerType.LeftFoot:
					dci.drive2XAngleTargetController1 = characterSelector.leftFootController;
					break;
				case DAZClothingItem.ControllerType.RightFoot:
					dci.drive2XAngleTargetController1 = characterSelector.rightFootController;
					break;
				case DAZClothingItem.ControllerType.LeftToe:
					dci.drive2XAngleTargetController1 = characterSelector.leftToeController;
					break;
				case DAZClothingItem.ControllerType.RightToe:
					dci.drive2XAngleTargetController1 = characterSelector.rightToeController;
					break;
			}
			
			switch (dci.driveXAngleTargetController2Type) {
				case DAZClothingItem.ControllerType.None:
					dci.driveXAngleTargetController2 = null;
					break;
				case DAZClothingItem.ControllerType.LeftFoot:
					dci.driveXAngleTargetController2 = characterSelector.leftFootController;
					break;
				case DAZClothingItem.ControllerType.RightFoot:
					dci.driveXAngleTargetController2 = characterSelector.rightFootController;
					break;
				case DAZClothingItem.ControllerType.LeftToe:
					dci.driveXAngleTargetController2 = characterSelector.leftToeController;
					break;
				case DAZClothingItem.ControllerType.RightToe:
					dci.driveXAngleTargetController2 = characterSelector.rightToeController;
					break;
			}

			switch (dci.drive2XAngleTargetController2Type) {
				case DAZClothingItem.ControllerType.None:
					dci.drive2XAngleTargetController2 = null;
					break;
				case DAZClothingItem.ControllerType.LeftFoot:
					dci.drive2XAngleTargetController2 = characterSelector.leftFootController;
					break;
				case DAZClothingItem.ControllerType.RightFoot:
					dci.drive2XAngleTargetController2 = characterSelector.rightFootController;
					break;
				case DAZClothingItem.ControllerType.LeftToe:
					dci.drive2XAngleTargetController2 = characterSelector.leftToeController;
					break;
				case DAZClothingItem.ControllerType.RightToe:
					dci.drive2XAngleTargetController2 = characterSelector.rightToeController;
					break;
			}
			RefreshDazClothingItem();
		}
		
		private void DisconnectClothingItem(DAZClothingItem shoeClothingItem){
			if(shoeClothingItem != null) {
				if(shoeClothingItem.driveXAngleTargetController1 != null) shoeClothingItem.driveXAngleTargetController1.jointRotationDriveXTarget = 0f;
				if(shoeClothingItem.drive2XAngleTargetController1 != null) shoeClothingItem.drive2XAngleTargetController1.jointRotationDriveXTarget = 0f;
			
				shoeClothingItem.driveXAngleTargetController1Type = DAZClothingItem.ControllerType.None;
				shoeClothingItem.drive2XAngleTargetController1Type = DAZClothingItem.ControllerType.None;
			

				if(shoeClothingItem.driveXAngleTargetController2 != null) shoeClothingItem.driveXAngleTargetController2.jointRotationDriveXTarget = 0f;
				if(shoeClothingItem.drive2XAngleTargetController2 != null) shoeClothingItem.drive2XAngleTargetController2.jointRotationDriveXTarget = 0f;
		
				shoeClothingItem.driveXAngleTargetController2Type = DAZClothingItem.ControllerType.None;
				shoeClothingItem.drive2XAngleTargetController2Type = DAZClothingItem.ControllerType.None;
				
				SyncDriveControllers(shoeClothingItem);
			}
			AdjustControllerRotation(0f);
			RefreshDazClothingItem();
		}
		
		public UIDynamicSlider AddSlider(string name, string label, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rightSide = false)
		{
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.label = label;
			uiSlider.rangeAdjustEnabled = true;
			RegisterFloat(storableFloat);
			
			storableFloat.setCallbackFunction = onChange;
			
			return uiSlider;
		}
		
		private void OnDisable()
		{
			if(shoeClothingItem != null){
				DisconnectClothingItem(shoeClothingItem);
			}
			foreach(var sc in shoeColliders) {
				sc.gameObject.SetActive(false);
			}
		}
		
		private void OnEnable()
		{
			if(shoeClothingItem != null){
				SyncShoeSettings(shoeClothingItem);
			}
		}
		
		private void OnDestroy(){
			foreach(var go in createdObjects) {
				UnityEngine.Object.Destroy(go);
			}
		}
	}
}
